# -*- coding: utf-8 -*-
"""
Tests for TestTcodDbExporter
"""
from django.utils import unittest

from aiida.backends.djsite.db.testbase import AiidaTestCase
from aiida.common.links import LinkType

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class FakeObject(object):
    """
    A wrapper for dictionary, which can be used instead of object.
    Example use case: fake Calculation object ``calc``, having keys
    ``inp`` and ``out`` to access also fake NodeInputManager and
    NodeOutputManager.
    """

    def __init__(self, dictionary):
        self._dictionary = dictionary

    def __getattr__(self, name):
        if isinstance(self._dictionary[name], dict):
            return FakeObject(self._dictionary[name])
        else:
            return self._dictionary[name]


class TestTcodDbExporter(AiidaTestCase):
    """
    Tests for TcodDbExporter class.
    """
    from aiida.orm.data.structure import has_ase, has_pyspglib
    from aiida.orm.data.cif import has_pycifrw

    def test_contents_encoding(self):
        """
        Testing the logic of choosing the encoding and the process of
        encoding contents.
        """
        from aiida.tools.dbexporters.tcod import cif_encode_contents
        self.assertEquals(cif_encode_contents('simple line')[1],
                          None)
        self.assertEquals(cif_encode_contents(' ;\n ;')[1],
                          None)
        self.assertEquals(cif_encode_contents(';\n'),
                          ('=3B\n', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('line\n;line'),
                          ('line\n=3Bline', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('tabbed\ttext'),
                          ('tabbed=09text', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('angstrom Å'),
                          ('angstrom =C3=85', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('.'),
                          ('=2E', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('?'),
                          ('=3F', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('.?'), ('.?', None))
        # This one is particularly tricky: a long line is folded by the QP
        # and the semicolon sign becomes the first character on a new line.
        self.assertEquals(cif_encode_contents(
            "Å{};a".format("".join("a" for i in range(0, 69)))),
            ('=C3=85aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
             'aaaaaaaaaaaaaaaaaaaaaaaaaaaaa=\n=3Ba',
             'quoted-printable'))
        self.assertEquals(cif_encode_contents('angstrom ÅÅÅ'),
                          ('YW5nc3Ryb20gw4XDhcOF', 'base64'))
        self.assertEquals(cif_encode_contents(
            "".join("a" for i in range(0, 2048)))[1],
                          None)
        self.assertEquals(cif_encode_contents(
            "".join("a" for i in range(0, 2049)))[1],
                          'quoted-printable')
        self.assertEquals(cif_encode_contents('datatest')[1], None)
        self.assertEquals(cif_encode_contents('data_test')[1], 'base64')

    def test_collect_files(self):
        """
        Testing the collection of files from file tree.
        """
        from aiida.tools.dbexporters.tcod import _collect_files
        from aiida.common.folders import SandboxFolder
        import StringIO

        sf = SandboxFolder()
        sf.get_subfolder('out', create=True)
        sf.get_subfolder('pseudo', create=True)
        sf.get_subfolder('save', create=True)
        sf.get_subfolder('save/1', create=True)
        sf.get_subfolder('save/2', create=True)

        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'aiida.in')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'aiida.out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, '_aiidasubmit.sh')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, '_.out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'out/out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'save/1/log.log')

        md5 = '098f6bcd4621d373cade4e832627b4f6'
        sha1 = 'a94a8fe5ccb19ba61c4c0873d391e987982fbbd3'
        self.assertEquals(
            _collect_files(sf.abspath),
            [{'name': '_.out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': '_aiidasubmit.sh', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'aiida.in', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'aiida.out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'out/', 'type': 'folder'},
             {'name': 'out/out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'pseudo/', 'type': 'folder'},
             {'name': 'save/', 'type': 'folder'},
             {'name': 'save/1/', 'type': 'folder'},
             {'name': 'save/1/log.log', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'save/2/', 'type': 'folder'}])

    @unittest.skipIf(not has_ase() or not has_pyspglib() or not has_pycifrw(),
                     "Unable to import ase or pyspglib")
    def test_cif_structure_roundtrip(self):
        from aiida.tools.dbexporters.tcod import export_cif, export_values
        from aiida.orm import Code
        from aiida.orm import JobCalculation
        from aiida.orm.data.cif import CifData
        from aiida.orm.data.parameter import ParameterData
        from aiida.orm.data.upf import UpfData
        from aiida.orm.data.folder import FolderData
        from aiida.common.folders import SandboxFolder
        from aiida.common.datastructures import calc_states
        import tempfile

        with tempfile.NamedTemporaryFile() as f:
            f.write('''
                data_test
                _cell_length_a    10
                _cell_length_b    10
                _cell_length_c    10
                _cell_angle_alpha 90
                _cell_angle_beta  90
                _cell_angle_gamma 90
                loop_
                _atom_site_label
                _atom_site_fract_x
                _atom_site_fract_y
                _atom_site_fract_z
                C 0 0 0
                O 0.5 0.5 0.5
            ''')
            f.flush()
            a = CifData(file=f.name)

        c = a._get_aiida_structure()
        c.store()
        pd = ParameterData()

        code = Code(local_executable='test.sh')
        with tempfile.NamedTemporaryFile() as f:
            f.write("#/bin/bash\n\necho test run\n")
            f.flush()
            code.add_path(f.name, 'test.sh')

        code.store()

        calc = JobCalculation(computer=self.computer)
        calc.set_resources({'num_machines': 1,
                            'num_mpiprocs_per_machine': 1})
        calc.add_link_from(code, "code")
        calc.set_environment_variables({'PATH': '/dev/null', 'USER': 'unknown'})

        with tempfile.NamedTemporaryFile(prefix="Fe") as f:
            f.write("<UPF version=\"2.0.1\">\nelement=\"Fe\"\n")
            f.flush()
            upf = UpfData(file=f.name)
            upf.store()
            calc.add_link_from(upf, "upf")

        with tempfile.NamedTemporaryFile() as f:
            f.write("data_test")
            f.flush()
            cif = CifData(file=f.name)
            cif.store()
            calc.add_link_from(cif, "cif")

        calc.store()
        calc._set_state(calc_states.SUBMITTING)
        with SandboxFolder() as f:
            calc._store_raw_input_folder(f.abspath)

        fd = FolderData()
        with open(fd._get_folder_pathsubfolder.get_abs_path(
                calc._SCHED_OUTPUT_FILE), 'w') as f:
            f.write("standard output")
            f.flush()

        with open(fd._get_folder_pathsubfolder.get_abs_path(
                calc._SCHED_ERROR_FILE), 'w') as f:
            f.write("standard error")
            f.flush()

        fd.store()
        fd.add_link_from(calc, calc._get_linkname_retrieved(), LinkType.CREATE)

        pd.add_link_from(calc, "calc", LinkType.CREATE)
        pd.store()

        with self.assertRaises(ValueError):
            export_cif(c, parameters=pd)

        c.add_link_from(calc, "calc", LinkType.CREATE)
        export_cif(c, parameters=pd)

        values = export_values(c, parameters=pd)
        values = values['0']

        self.assertEquals(values['_tcod_computation_environment'],
                          ['PATH=/dev/null\nUSER=unknown'])
        self.assertEquals(values['_tcod_computation_command'],
                          ['cd 0; ./_aiidasubmit.sh'])

    def test_pw_translation(self):
        from aiida.tools.dbexporters.tcod \
            import translate_calculation_specific_values
        from aiida.tools.dbexporters.tcod_plugins.pw \
            import PwTcodtranslator as PWT
        from aiida.tools.dbexporters.tcod_plugins.cp \
            import CpTcodtranslator as CPT
        from aiida.orm.code import Code
        from aiida.orm.data.array import ArrayData
        from aiida.orm.data.array.kpoints import KpointsData
        from aiida.orm.data.parameter import ParameterData
        import numpy

        code = Code()
        code._set_attr('remote_exec_path', '/test')

        kpoints = KpointsData()
        kpoints.set_kpoints_mesh([2, 3, 4], offset=[0.25, 0.5, 0.75])

        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={}),
                    "kpoints": kpoints, "code": code},
            "out": {"output_parameters": ParameterData(dict={})}
        })
        res = translate_calculation_specific_values(calc, PWT)
        self.assertEquals(res, {
            '_integration_grid_X': 2,
            '_integration_grid_Y': 3,
            '_integration_grid_Z': 4,
            '_integration_grid_shift_X': 0.25,
            '_integration_grid_shift_Y': 0.5,
            '_integration_grid_shift_Z': 0.75,
            '_tcod_software_package': 'Quantum ESPRESSO',
            '_tcod_software_executable_path': '/test',
        })

        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={
                'SYSTEM': {'ecutwfc': 40, 'occupations': 'smearing'}
            })},
            "out": {"output_parameters": ParameterData(dict={
                'number_of_electrons': 10,
            })}
        })
        res = translate_calculation_specific_values(calc, PWT)
        self.assertEquals(res, {
            '_dft_cell_valence_electrons': 10,
            '_tcod_software_package': 'Quantum ESPRESSO',
            '_dft_BZ_integration_smearing_method': 'Gaussian',
            '_dft_kinetic_energy_cutoff_EEX': 2176.910676048,
            '_dft_kinetic_energy_cutoff_charge_density': 2176.910676048,
            '_dft_kinetic_energy_cutoff_wavefunctions': 544.227669012,
        })

        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={})},
            "out": {"output_parameters": ParameterData(dict={
                'energy_xc': 5,
            })}
        })
        with self.assertRaises(ValueError):
            translate_calculation_specific_values(calc, PWT)

        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={})},
            "out": {"output_parameters": ParameterData(dict={
                'energy_xc': 5,
                'energy_xc_units': 'meV'
            })}
        })
        with self.assertRaises(ValueError):
            translate_calculation_specific_values(calc, PWT)

        energies = {
            'energy': -3701.7004199449257,
            'energy_one_electron': -984.0078459766,
            'energy_xc': -706.6986753641559,
            'energy_ewald': -2822.6335103043157,
            'energy_hartree': 811.6396117001462,
            'fermi_energy': 10.25208617898623,
        }
        dct = energies
        for key in energies.keys():
            dct["{}_units".format(key)] = 'eV'
        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={
                'SYSTEM': {'smearing': 'mp'}
            })},
            "out": {"output_parameters": ParameterData(dict=dct)}
        })
        res = translate_calculation_specific_values(calc, PWT)
        self.assertEquals(res, {
            '_tcod_total_energy': energies['energy'],
            '_dft_1e_energy': energies['energy_one_electron'],
            '_dft_correlation_energy': energies['energy_xc'],
            '_dft_ewald_energy': energies['energy_ewald'],
            '_dft_hartree_energy': energies['energy_hartree'],
            '_dft_fermi_energy': energies['fermi_energy'],
            '_tcod_software_package': 'Quantum ESPRESSO',
            '_dft_BZ_integration_smearing_method': 'Methfessel-Paxton',
            '_dft_BZ_integration_MP_order': 1,
        })
        dct = energies
        dct['number_of_electrons'] = 10
        for key in energies.keys():
            dct["{}_units".format(key)] = 'eV'
        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={
                'SYSTEM': {'smearing': 'unknown-method'}
            })},
            "out": {"output_parameters": ParameterData(dict=dct)}
        })
        res = translate_calculation_specific_values(calc, CPT)
        self.assertEquals(res, {'_dft_cell_valence_electrons': 10,
                                '_tcod_software_package':
                                    'Quantum ESPRESSO'})

        ad = ArrayData()
        ad.set_array("forces", numpy.array([[[1, 2, 3], [4, 5, 6]]]))
        calc = FakeObject({
            "inp": {"parameters": ParameterData(dict={
                'SYSTEM': {'smearing': 'unknown-method'}
            })},
            "out": {"output_parameters": ParameterData(dict={}),
                    "output_array": ad}
        })
        res = translate_calculation_specific_values(calc, PWT)
        self.assertEquals(res, {
            '_tcod_software_package': 'Quantum ESPRESSO',
            '_dft_BZ_integration_smearing_method': 'other',
            '_dft_BZ_integration_smearing_method_other': 'unknown-method',
            ## Residual forces are no longer produced, as they should
            ## be in the same CIF loop with coordinates -- to be
            ## implemented later, since it's not yet clear how.
            # '_tcod_atom_site_resid_force_Cartn_x': [1,4],
            # '_tcod_atom_site_resid_force_Cartn_y': [2,5],
            # '_tcod_atom_site_resid_force_Cartn_z': [3,6],
        })

    def test_nwcpymatgen_translation(self):
        from aiida.tools.dbexporters.tcod \
            import translate_calculation_specific_values
        from aiida.tools.dbexporters.tcod_plugins.nwcpymatgen \
            import NwcpymatgenTcodtranslator as NPT
        from aiida.orm.data.parameter import ParameterData
        from tcodexporter import FakeObject

        calc = FakeObject({
            "out": {"output":
                ParameterData(dict={
                    "basis_set": {
                        "H": {
                            "description": "6-31g",
                            "functions": "2",
                            "shells": "2",
                            "types": "2s"
                        },
                        "O": {
                            "description": "6-31g",
                            "functions": "9",
                            "shells": "5",
                            "types": "3s2p"
                        }
                    },
                    "corrections": {},
                    "energies": [
                        -2057.99011937535
                    ],
                    "errors": [],
                    "frequencies": None,
                    "has_error": False,
                    "job_type": "NWChem SCF Module"
                }),
                "job_info": ParameterData(dict={
                    "0 permanent": ".",
                    "0 scratch": ".",
                    "argument  1": "aiida.in",
                    "compiled": "Sun_Dec_22_04:02:59_2013",
                    "data base": "./aiida.db",
                    "date": "Mon May 11 17:10:07 2015",
                    "ga revision": "10379",
                    "global": "200.0 Mbytes (distinct from heap & stack)",
                    "hardfail": "no",
                    "heap": "100.0 Mbytes",
                    "hostname": "theospc11",
                    "input": "aiida.in",
                    "nproc": "6",
                    "nwchem branch": "6.3",
                    "nwchem revision": "24277",
                    "prefix": "aiida.",
                    "program": "/usr/bin/nwchem",
                    "source": "/build/buildd/nwchem-6.3+r1",
                    "stack": "100.0 Mbytes",
                    "status": "startup",
                    "time left": "-1s",
                    "total": "400.0 Mbytes",
                    "verify": "yes",
                })
            }})
        res = translate_calculation_specific_values(calc, NPT)
        self.assertEquals(res, {
            '_tcod_software_package': 'NWChem',
            '_tcod_software_package_version': '6.3',
            '_tcod_software_package_compilation_date': '2013-12-22T04:02:59',
            '_atom_type_symbol': ['H', 'O'],
            '_dft_atom_basisset': ['6-31g', '6-31g'],
            '_dft_atom_type_valence_configuration': ['2s', '3s2p'],
        })

    @unittest.skipIf(not has_ase() or not has_pycifrw() or not has_pyspglib(),
                     "Unable to import ase, pycifrw or pyspglib")
    def test_inline_export(self):
        from aiida.orm.data.cif import CifData
        from aiida.tools.dbexporters.tcod import export_values
        import tempfile

        with tempfile.NamedTemporaryFile() as f:
            f.write('''
                data_test
                _cell_length_a    10
                _cell_length_b    10
                _cell_length_c    10
                _cell_angle_alpha 90
                _cell_angle_beta  90
                _cell_angle_gamma 90
                loop_
                _atom_site_label
                _atom_site_fract_x
                _atom_site_fract_y
                _atom_site_fract_z
                C 0 0 0
                O 0.5 0.5 0.5
            ''')
            f.flush()
            a = CifData(file=f.name)

        s = a._get_aiida_structure(store=True)
        val = export_values(s)
        script = val.first_block()['_tcod_file_contents'][1]
        function = '_get_aiida_structure_ase_inline'
        self.assertNotEqual(script.find(function), script.rfind(function))

    @unittest.skipIf(not has_ase() or not has_pyspglib() or not has_pycifrw(),
                     "Unable to import ase or pyspglib")
    def test_symmetry_reduction(self):
        from aiida.orm.data.structure import StructureData
        from aiida.tools.dbexporters.tcod import export_values
        from ase import Atoms

        a = Atoms('BaTiO3', cell=(4., 4., 4.))
        a.set_scaled_positions(
            ((0.0, 0.0, 0.0),
             (0.5, 0.5, 0.5),
             (0.5, 0.5, 0.0),
             (0.5, 0.0, 0.5),
             (0.0, 0.5, 0.5),
             )
        )

        a.set_chemical_symbols(['Ba', 'Ti', 'O', 'O', 'O'])
        val = export_values(StructureData(ase=a), reduce_symmetry=True, store=True)['0']
        self.assertEqual(val['_atom_site_label'], ['Ba1', 'Ti1', 'O1'])
        self.assertEqual(val['_symmetry_space_group_name_H-M'], 'Pm-3m')
        self.assertEqual(val['_symmetry_space_group_name_Hall'], '-P 4 2 3')

    def test_cmdline_parameters(self):
        """
        Ensuring that neither extend_with_cmdline_parameters() nor
        deposition_cmdline_parameters() set default parameters.
        """
        from aiida.tools.dbexporters.tcod \
            import extend_with_cmdline_parameters, \
            deposition_cmdline_parameters
        import argparse

        parser = argparse.ArgumentParser()
        extend_with_cmdline_parameters(parser)
        options = vars(parser.parse_args(args=[]))

        for key in options.keys():
            if options[key] is None:
                options.pop(key)

        self.assertEqual(options, {})

        parser = argparse.ArgumentParser()
        deposition_cmdline_parameters(parser)
        options = vars(parser.parse_args(args=[]))

        for key in options.keys():
            if options[key] is None:
                options.pop(key)

        self.assertEqual(options, {})

    @unittest.skipIf(not has_ase() or not has_pycifrw() or not has_pyspglib(),
                     "Unable to import ase, pycifrw or pyspglib")
    def test_export_trajectory(self):
        from aiida.orm.data.structure import StructureData
        from aiida.orm.data.array.trajectory import TrajectoryData
        from aiida.tools.dbexporters.tcod import export_values

        cells = [
            [[2., 0., 0., ],
             [0., 2., 0., ],
             [0., 0., 2., ]],
            [[3., 0., 0., ],
             [0., 3., 0., ],
             [0., 0., 3., ]]
        ]
        symbols = [['H', 'O', 'C'], ['H', 'O', 'C']]
        positions = [
            [[0., 0., 0.],
             [0.5, 0.5, 0.5],
             [1.5, 1.5, 1.5]],
            [[0., 0., 0.],
             [0.75, 0.75, 0.75],
             [1.25, 1.25, 1.25]]
        ]
        structurelist = []
        for i in range(0, 2):
            struct = StructureData(cell=cells[i])
            for j, symbol in enumerate(symbols[i]):
                struct.append_atom(symbols=symbol, position=positions[i][j])
            structurelist.append(struct)

        td = TrajectoryData(structurelist=structurelist)

        with self.assertRaises(ValueError):
            # Trajectory index is not specified
            v = export_values(td)

        expected_tags = [
            '_atom_site_fract_x',
            '_atom_site_fract_y',
            '_atom_site_fract_z',
            '_atom_site_label',
            '_atom_site_type_symbol',
            '_audit_conform_dict_location',
            '_audit_conform_dict_name',
            '_audit_conform_dict_version',
            '_audit_creation_method',
            '_cell_angle_alpha',
            '_cell_angle_beta',
            '_cell_angle_gamma',
            '_cell_length_a',
            '_cell_length_b',
            '_cell_length_c',
            '_chemical_formula_sum',
            '_symmetry_Int_Tables_number',
            '_symmetry_equiv_pos_as_xyz',
            '_symmetry_space_group_name_H-M',
            '_symmetry_space_group_name_Hall'
        ]

        tcod_file_tags = [
            '_tcod_content_encoding_id',
            '_tcod_content_encoding_layer_id',
            '_tcod_content_encoding_layer_type',
            '_tcod_file_URI',
            '_tcod_file_content_encoding',
            '_tcod_file_contents',
            '_tcod_file_id',
            '_tcod_file_md5sum',
            '_tcod_file_name',
            '_tcod_file_role',
            '_tcod_file_sha1sum'
        ]

        # Not stored and not to be stored:
        v = export_values(td, trajectory_index=1)
        self.assertEqual(sorted(v['0'].keys()), expected_tags)

        # Stored, but not expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)

        # Not stored, but expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        v = export_values(td, trajectory_index=1, store=True)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)

        # Both stored and expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1, store=True)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)

        # Stored, but asked not to include DB dump:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1,
                          dump_aiida_database=False)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags)

    def test_contents_encoding(self):
        """
        Testing the logic of choosing the encoding and the process of
        encoding contents.
        """
        from aiida.tools.dbexporters.tcod import decode_textfield

        def test_ncr(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_ncr,
                                                      decode_textfield_ncr)
            encoded = encode_textfield_ncr(inp)
            decoded = decode_textfield_ncr(out)
            decoded_universal = decode_textfield(out, 'ncr')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def test_quoted_printable(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_quoted_printable,
                                                      decode_textfield_quoted_printable)
            encoded = encode_textfield_quoted_printable(inp)
            decoded = decode_textfield_quoted_printable(out)
            decoded_universal = decode_textfield(out, 'quoted-printable')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def test_base64(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_base64,
                                                      decode_textfield_base64)
            encoded = encode_textfield_base64(inp)
            decoded = decode_textfield_base64(out)
            decoded_universal = decode_textfield(out, 'base64')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def test_gzip_base64(self, text):
            from aiida.tools.dbexporters.tcod import (encode_textfield_gzip_base64,
                                                      decode_textfield_gzip_base64)
            encoded = encode_textfield_gzip_base64(text)
            decoded = decode_textfield_gzip_base64(encoded)
            decoded_universal = decode_textfield(encoded, 'gzip+base64')
            self.assertEquals(text, decoded)
            self.assertEquals(text, decoded_universal)

        test_ncr(self, '.', '&#46;')
        test_ncr(self, '?', '&#63;')
        test_ncr(self, ';\n', '&#59;\n')
        test_ncr(self, 'line\n;line', 'line\n&#59;line')
        test_ncr(self, 'tabbed\ttext', 'tabbed&#9;text')
        test_ncr(self, 'angstrom Å', 'angstrom &#195;&#133;')
        test_ncr(self, '<html>&#195;&#133;</html>',
                 '<html>&#38;#195;&#38;#133;</html>')

        test_quoted_printable(self, '.', '=2E')
        test_quoted_printable(self, '?', '=3F')
        test_quoted_printable(self, ';\n', '=3B\n')
        test_quoted_printable(self, 'line\n;line', 'line\n=3Bline')
        test_quoted_printable(self, 'tabbed\ttext', 'tabbed=09text')
        test_quoted_printable(self, 'angstrom Å', 'angstrom =C3=85')
        test_quoted_printable(self, 'line\rline\x00', 'line\rline=00')
        # This one is particularly tricky: a long line is folded by the QP
        # and the semicolon sign becomes the first character on a new line.
        test_quoted_printable(self,
                              "Å{};a".format("".join("a" for i in range(0, 69))),
                              '=C3=85aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
                              'aaaaaaaaaaaaaaaaaaaaaaaaaaaaa=\n=3Ba')

        test_base64(self, 'angstrom ÅÅÅ', 'YW5nc3Ryb20gw4XDhcOF')
        test_gzip_base64(self, 'angstrom ÅÅÅ')



# -*- coding: utf-8 -*-

import argparse
import sys

from aiida.backends.utils import load_dbenv, is_dbenv_loaded
from aiida.cmdline.baseclass import VerdiCommandWithSubcommands
from aiida.common.exceptions import NotExistent

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class Graph(VerdiCommandWithSubcommands):
    """
    Utility to explore the nodes in the database graph
    More specifically it allow the user to:
        1) Generate and render graph
        2) (TODO) Find if two nodes of a graph are connected
        3) (TODO) Extend functionalities

    A list dictionary with available subcommands can be found in __init__ function
    """

    def __init__(self):
        """
        A dictionary with valid subcommandsa as keys and corresponding functions as values
        """
        self.valid_subcommands = {
            'generate': (self.graph_generate, self.complete_none)
            # TODO: add a command to find connections between two points
        }

    def graph_generate(self, *args):
        """
        Function to generate a graph given a root node user-specified by its pk.
        :param args: root_pk
        :return: Generate a .dot file that can be rendered by graphviz utility dot
        """

        # Usual boilerplate to load certain classes and functions
        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm import load_node
        from aiida.orm.calculation import Calculation
        from aiida.orm.calculation.job import JobCalculation
        from aiida.orm.code import Code
        from aiida.orm.data.array.kpoints import KpointsData
        from aiida.orm.data.structure import StructureData
        from aiida.orm.calculation.inline import InlineCalculation

        # Parse input arguments
        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Generate a graph from a root node')

        parser.add_argument('ROOT', help="The pk of the root node",
                            type=int)

        # Parse args and retrieve root pk
        args = list(args)
        parsed_args = parser.parse_args(args)
        root_pk = parsed_args.ROOT

        # Try to retrieve the node
        try:
            n = load_node(root_pk)
        except NotExistent as e:
            print >> sys.stderr, e.message
            sys.exit(1)

        # From this point on I inherit Giovanni's script grap-generator.py.
        # This script starts from the root_pk and goes both input-ward and output-ward via a breadth-first algorithm
        # until the connected part of the graph that contains the root_pk is fully explored.
        # TODO this command deserves to be radically improved, with options and further subcommands


        def kpoints_desc(node):
            """
            Returns a string with infos retrieved from  kpoints node's properties.
            :param node:
            :return: retstr
            """
            try:
                mesh = node.get_kpoints_mesh()
                return "{}x{}x{} (+{:.1f},{:.1f},{:.1f})".format(
                    mesh[0][0], mesh[0][1], mesh[0][2],
                    mesh[1][0], mesh[1][1], mesh[1][2])
            except AttributeError:
                return '({} kpts)'.format(len(node.get_kpoints()))

        def pw_desc(node):
            """
            Returns a string with infos retrieved from  PwCalculation node's properties.
            :param node:
            :return: retsrt:
            """
            ############I won't use this until node.inp. methods won't be working correctly
            # return '{}'.format(node.inp.parameters.dict.CONTROL['calculation'])
            ###########

            ###############3I rather use this black magic
            return '{}'.format(dict(node.get_inputs(also_labels=True))['parameters'].dict.CONTROL['calculation'])

        def get_additional_string(node):
            """
            Returns a string with infos retrieved from  PwCalculation node's properties.
            The actual returned string depends on the node class
            :param node:
            :return: retstr
            """
            class_name = node.__class__.__name__

            func_mapping = {
                "StructureData": lambda x: x.get_formula(mode='hill_compact'),
                "InlineCalculation": lambda x: "{}()".format(x.get_function_name()),
                "KpointsData": kpoints_desc,
                "PwCalculation": pw_desc,
            }

            func = func_mapping.get(class_name, None)
            if func is None:
                retstr = ""

            else:
                retstr = "{}".format(func(node))

            if isinstance(node, JobCalculation):
                retstr = " ".join([retstr, node.get_state(from_attribute=True)])

            if retstr:
                return "\n{}".format(retstr)
            else:
                return ""

        def draw_node_settings(node, **kwargs):
            """
            Returns a string with all infos needed in a .dot file  to define a node of a graph.
            :param node:
            :param kwargs: Additional key-value pairs to be added to the returned string
            :return: a string
            """
            if isinstance(node, Calculation):
                shape = "shape=polygon,sides=4"
            elif isinstance(node, Code):
                shape = "shape=diamond"
            else:
                shape = "shape=ellipse"
            if kwargs:
                additional_params = ",{}".format(
                    ",".join('{}="{}"'.format(k, v) for k, v in kwargs.iteritems()))
            else:
                additional_params = ""
            additional_string = get_additional_string(node)
            if node.label:
                label_string = "\n'{}'".format(node.label)
            else:
                label_string = ""
            labelstring = 'label="{} ({}){}{}"'.format(
                node.__class__.__name__, node.pk, label_string,
                additional_string)
            return "N{} [{},{}{}];".format(node.pk, shape, labelstring,
                                           additional_params)


        # Breadth-first search of all ancestors and descendant nodes of a given node
        links = []  # Accumulate links here
        nodes = {n.pk: draw_node_settings(n, style='filled', color='lightblue')} #Accumulate nodes specs here

        last_nodes = [n] # Put the nodes whose links have not been scanned yet

        # Go through the graph on-ward (i.e. look at inputs)
        while last_nodes:
            new_nodes = []
            for node in last_nodes:
                inputs = node.get_inputs(also_labels=True)
                for linkname, inp in inputs:
                    links.append((inp.pk, node.pk, linkname))
                    if inp.pk not in nodes:
                        nodes[inp.pk] = draw_node_settings(inp)
                        new_nodes.append(inp)
            last_nodes = new_nodes

        # Go through the graph down-ward (i.e. look at outputs)
        last_nodes = [n]
        while last_nodes:
            new_nodes = []
            for node in last_nodes:
                outputs = node.get_outputs(also_labels=True)
                for linkname, out in outputs:
                    links.append((node.pk, out.pk, linkname))
                    if out.pk not in nodes:
                        nodes[out.pk] = draw_node_settings(out)
                        new_nodes.append(out)
            last_nodes = new_nodes

        # Generate name of the output file. Default
        out_file_name = "{}.dot".format(root_pk)
        with open(out_file_name,'w') as fout:

            fout.write("digraph G {\n")

            for l in links:
                fout.write('    {} -> {} [label="{}"];\n'.format("N{}".format(l[0]),  "N{}".format(l[1]), l[2]))
            for n_name, n_values in nodes.iteritems():
                fout.write("    {}\n".format(n_values))

            fout.write("}\n")



# -*- coding: utf-8 -*-

import collections

from aiida.orm.implementation.general.group import AbstractGroup

from aiida.common.exceptions import (ModificationNotAllowed, UniquenessError,
                                     NotExistent)

from aiida.orm.implementation.django.node import Node

from aiida.backends.djsite.utils import get_automatic_user

from django.db import transaction, IntegrityError
from django.db.models import Q
from django.core.exceptions import ObjectDoesNotExist

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"


class Group(AbstractGroup):
    def __init__(self, **kwargs):
        from aiida.backends.djsite.db.models import DbGroup
        dbgroup = kwargs.pop('dbgroup', None)

        if dbgroup is not None:
            if isinstance(dbgroup, (int, long)):
                try:
                    dbgroup = DbGroup.objects.get(pk=dbgroup)
                except ObjectDoesNotExist:
                    raise NotExistent("Group with pk={} does not exist".format(
                        dbgroup))
            if not isinstance(dbgroup, DbGroup):
                raise TypeError("dbgroup is not a DbGroup instance, it is "
                                "instead {}".format(str(type(dbgroup))))
            if kwargs:
                raise ValueError("If you pass a dbgroups, you cannot pass any "
                                 "further parameter")

            self._dbgroup = dbgroup
        else:
            name = kwargs.pop('name', None)
            if name is None:
                raise ValueError("You have to specify a group name")
            group_type = kwargs.pop('type_string',
                                    "")  # By default, an user group
            user = kwargs.pop('user', get_automatic_user())
            description = kwargs.pop('description', "")
            self._dbgroup = DbGroup(name=name, description=description,
                                    user=user, type=group_type)
            if kwargs:
                raise ValueError("Too many parameters passed to Group, the "
                                 "unknown parameters are: {}".format(
                    ", ".join(kwargs.keys())))

    @property
    def name(self):
        return self.dbgroup.name

    @property
    def description(self):
        return self.dbgroup.description

    @description.setter
    def description(self, value):
        self.dbgroup.description = value

        # Update the entry in the DB, if the group is already stored
        if self.is_stored:
            self.dbgroup.save()

    @property
    def type_string(self):
        return self.dbgroup.type

    @property
    def user(self):
        return self.dbgroup.user

    @property
    def dbgroup(self):
        return self._dbgroup

    @property
    def pk(self):
        return self._dbgroup.pk

    @property
    def id(self):
        return self._dbgroup.pk

    @property
    def uuid(self):
        return unicode(self.dbgroup.uuid)

    def __int__(self):
        if self._to_be_stored:
            return None
        else:
            return self._dbnode.pk

    @property
    def is_stored(self):
        return self.pk is not None

    def store(self):
        if not self.is_stored:
            sid = transaction.savepoint()
            try:
                self.dbgroup.save()
            except IntegrityError:
                transaction.savepoint_rollback(sid)
                raise UniquenessError("A group with the same name (and of the "
                                      "same type) already "
                                      "exists, unable to store")

        # To allow to do directly g = Group(...).store()
        return self

    def add_nodes(self, nodes):
        from aiida.backends.djsite.db.models import DbNode
        if not self.is_stored:
            raise ModificationNotAllowed("Cannot add nodes to a group before "
                                         "storing")

        # First convert to a list
        if isinstance(nodes, (Node, DbNode)):
            nodes = [nodes]

        if isinstance(nodes, basestring) or not isinstance(
                nodes, collections.Iterable):
            raise TypeError("Invalid type passed as the 'nodes' parameter to "
                            "add_nodes, can only be a Node, DbNode, or a list "
                            "of such objects, it is instead {}".format(
                str(type(nodes))))

        list_pk = []
        for node in nodes:
            if not isinstance(node, (Node, DbNode)):
                raise TypeError("Invalid type of one of the elements passed "
                                "to add_nodes, it should be either a Node or "
                                "a DbNode, it is instead {}".format(
                    str(type(node))))
            if node.pk is None:
                raise ValueError("At least one of the provided nodes is "
                                 "unstored, stopping...")
            list_pk.append(node.pk)

        self.dbgroup.dbnodes.add(*list_pk)

    @property
    def nodes(self):
        class iterator(object):
            def __init__(self, dbnodes):
                self.dbnodes = dbnodes
                self.generator = self._genfunction()

            def _genfunction(self):
                for n in self.dbnodes:
                    yield n.get_aiida_class()

            def __iter__(self):
                return self

            def __len__(self):
                return self.dbnodes.count()

            # For future python-3 compatibility
            def __next__(self):
                return self.next()

            def next(self):
                return next(self.generator)

        return iterator(self.dbgroup.dbnodes.all())

    def remove_nodes(self, nodes):
        from aiida.backends.djsite.db.models import DbNode
        if not self.is_stored:
            raise ModificationNotAllowed("Cannot remove nodes from a group "
                                         "before storing")

        # First convert to a list
        if isinstance(nodes, (Node, DbNode)):
            nodes = [nodes]

        if isinstance(nodes, basestring) or not isinstance(
                nodes, collections.Iterable):
            raise TypeError("Invalid type passed as the 'nodes' parameter to "
                            "remove_nodes, can only be a Node, DbNode, or a "
                            "list of such objects, it is instead {}".format(
                str(type(nodes))))

        list_pk = []
        for node in nodes:
            if not isinstance(node, (Node, DbNode)):
                raise TypeError("Invalid type of one of the elements passed "
                                "to add_nodes, it should be either a Node or "
                                "a DbNode, it is instead {}".format(
                    str(type(node))))
            if node.pk is None:
                raise ValueError("At least one of the provided nodes is "
                                 "unstored, stopping...")
            list_pk.append(node.pk)

        self.dbgroup.dbnodes.remove(*list_pk)

    @classmethod
    def query(cls, name=None, type_string="", pk=None, uuid=None, nodes=None,
              user=None, node_attributes=None, past_days=None,
              name_filters=None, **kwargs):

        from aiida.backends.djsite.db.models import (DbGroup, DbNode,
                                                     DbAttribute)

        # Analyze args and kwargs to create the query
        queryobject = Q()
        if name is not None:
            queryobject &= Q(name=name)

        if type_string is not None:
            queryobject &= Q(type=type_string)

        if pk is not None:
            queryobject &= Q(pk=pk)

        if uuid is not None:
            queryobject &= Q(uuid=uuid)

        if past_days is not None:
            queryobject &= Q(time__gte=past_days)

        if nodes is not None:
            pk_list = []

            if not isinstance(nodes, collections.Iterable):
                nodes = [nodes]

            for node in nodes:
                if not isinstance(node, (Node, DbNode)):
                    raise TypeError("At least one of the elements passed as "
                                    "nodes for the query on Group is neither "
                                    "a Node nor a DbNode")
                pk_list.append(node.pk)

            queryobject &= Q(dbnodes__in=pk_list)

        if user is not None:
            if isinstance(user, basestring):
                queryobject &= Q(user__email=user)
            else:
                queryobject &= Q(user=user)

        if name_filters is not None:
            name_filters_list = {"name__" + k: v for (k, v)
                                 in name_filters.iteritems() if v}
            queryobject &= Q(**name_filters_list)

        groups_pk = set(DbGroup.objects.filter(
            queryobject, **kwargs).values_list('pk', flat=True))

        if node_attributes is not None:
            for k, vlist in node_attributes.iteritems():
                if isinstance(vlist, basestring) or not isinstance(
                        vlist, collections.Iterable):
                    vlist = [vlist]

                for v in vlist:
                    # This will be a dictionary of the type
                    # {'datatype': 'txt', 'tval': 'xxx') for instance, if
                    # the passed data is a string
                    base_query_dict = DbAttribute.get_query_dict(v)
                    # prepend to the key the right django string to SQL-join
                    # on the right table
                    query_dict = {'dbnodes__dbattributes__{}'.format(k2): v2
                                  for k2, v2 in base_query_dict.iteritems()}

                    # I narrow down the list of groups.
                    # I had to do it in this way, with multiple DB hits and
                    # not a single, complicated query because in SQLite
                    # there is a maximum of 64 tables in a join.
                    # Since typically one requires a small number of filters,
                    # this should be ok.
                    groups_pk = groups_pk.intersection(DbGroup.objects.filter(
                        pk__in=groups_pk, dbnodes__dbattributes__key=k,
                        **query_dict).values_list('pk', flat=True))

        retlist = []
        # Return sorted by pk
        for dbgroup in sorted(groups_pk):
            retlist.append(cls(dbgroup=dbgroup))

        return retlist

    def delete(self):
        if self.pk is not None:
            self.dbgroup.delete()



# -*- coding: utf-8 -*-
"""
Plugin to create a Quantum Espresso pw.x file.

TODO: COPY OUTDIR FROM PREVIOUS CALCULATION! Should be an input node of type
     RemoteData (or maybe subclass it?).
TODO: tests!
TODO: DOC + implementation of SETTINGS
TODO: preexec, postexec
TODO: Check that no further parameters are passed in SETTINGS
TODO: many cards missing: check and implement
      e.g.: ['CONSTRAINTS', 'OCCUPATIONS']
TODO: implement pre_... and post_... hooks to add arbitrary strings before
      and after a namelist, and a 'final_string' (all optional); useful 
      for development when new cards are needed
TODO: all a lot of logger.debug stuff
"""
import os

from aiida.orm.calculation.job import JobCalculation
from aiida.orm.calculation.job.quantumespresso import BasePwCpInputGenerator
from aiida.common.utils import classproperty

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class CpCalculation(BasePwCpInputGenerator, JobCalculation):
    """
    Car-Parrinello molecular dynamics code (cp.x) of the
    Quantum ESPRESSO distribution.
    For more information, refer to http://www.quantum-espresso.org/
    """

    def _init_internal_params(self):
        super(CpCalculation, self)._init_internal_params()

        _cp_read_unit_number = 50
        _cp_write_unit_number = 51

        self._DATAFILE_XML = os.path.join(
            BasePwCpInputGenerator._OUTPUT_SUBFOLDER,
            '{}_{}.save'.format(BasePwCpInputGenerator._PREFIX,
                                _cp_write_unit_number),
            BasePwCpInputGenerator._DATAFILE_XML_BASENAME)

        self._FILE_XML_PRINT_COUNTER = os.path.join(
            BasePwCpInputGenerator._OUTPUT_SUBFOLDER,
            '{}_{}.save'.format(BasePwCpInputGenerator._PREFIX,
                                _cp_write_unit_number),
            self._FILE_XML_PRINT_COUNTER_BASENAME)

        # Default output parser provided by AiiDA
        self._default_parser = 'quantumespresso.cp'

        self._automatic_namelists = {
            'scf': ['CONTROL', 'SYSTEM', 'ELECTRONS'],
            'nscf': ['CONTROL', 'SYSTEM', 'ELECTRONS'],
            'relax': ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS'],
            'cp': ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS'],
            'vc-cp': ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL'],
            'vc-relax': ['CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL'],
            'vc-wf': ['CONTROL', 'SYSTEM', 'ELECTRONS', 'WANNIER'],
        }

        # Keywords that cannot be set
        self._blocked_keywords = [('CONTROL', 'pseudo_dir'),  # set later
                                  ('CONTROL', 'outdir'),  # set later
                                  ('CONTROL', 'prefix'),  # set later
                                  ('SYSTEM', 'ibrav'),  # set later
                                  ('SYSTEM', 'celldm'),
                                  ('SYSTEM', 'nat'),  # set later
                                  ('SYSTEM', 'ntyp'),  # set later
                                  ('SYSTEM', 'a'), ('SYSTEM', 'b'), ('SYSTEM', 'c'),
                                  ('SYSTEM', 'cosab'), ('SYSTEM', 'cosac'), ('SYSTEM', 'cosbc'),
                                  ('CONTROL', 'ndr', _cp_read_unit_number),
                                  ('CONTROL', 'ndw', _cp_write_unit_number),
        ]

        self._use_kpoints = False

        # in restarts, it will copy from the parent the following 
        self._restart_copy_from = os.path.join(
            BasePwCpInputGenerator._OUTPUT_SUBFOLDER,
            '{}_{}.save'.format(BasePwCpInputGenerator._PREFIX,
                                _cp_write_unit_number))
        # in restarts, it will copy the previous folder in the following one 
        self._restart_copy_to = os.path.join(
            BasePwCpInputGenerator._OUTPUT_SUBFOLDER,
            '{}_{}.save'.format(BasePwCpInputGenerator._PREFIX,
                                _cp_read_unit_number))

        _cp_ext_list = ['cel', 'con', 'eig', 'evp', 'for', 'nos', 'pol',
                        'pos', 'spr', 'str', 'the', 'vel', 'wfc']

        # I retrieve them all, even if I don't parse all of them
        self._internal_retrieve_list = [os.path.join(
            BasePwCpInputGenerator._OUTPUT_SUBFOLDER,
            '{}.{}'.format(BasePwCpInputGenerator._PREFIX,
                           ext)) for ext in _cp_ext_list]
        self._internal_retrieve_list += [self._FILE_XML_PRINT_COUNTER]

        # Default input and output files
        self._DEFAULT_INPUT_FILE = 'aiida.in'
        self._DEFAULT_OUTPUT_FILE = 'aiida.out'

    @classproperty
    def _FILE_XML_PRINT_COUNTER_BASENAME(cls):
        return 'print_counter.xml'

    @classproperty
    def _default_verbosity(cls):
        return 'low'


    @classproperty
    def _use_methods(cls):
        """
        Extend the parent _use_methods with further keys.
        """
        retdict = JobCalculation._use_methods
        retdict.update(BasePwCpInputGenerator._baseclass_use_methods)

        return retdict
    
    



# -*- coding: utf-8 -*-
import unittest

from aiida.transport.plugins.local import *

# This will be used by test_all_plugins

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

plugin_transport = LocalTransport()


class TestGeneric(unittest.TestCase):
    """
    Test whoami on localhost.
    """

    def test_whoami(self):
        import getpass

        with LocalTransport() as t:
            self.assertEquals(t.whoami(), getpass.getuser())


class TestBasicConnection(unittest.TestCase):
    """
    Test basic connections.
    """

    def test_closed_connection(self):
        from aiida.transport import TransportInternalError

        with self.assertRaises(TransportInternalError):
            t = LocalTransport()
            t.listdir()

    def test_invalid_param(self):
        with self.assertRaises(ValueError):
            LocalTransport(unrequired_var='something')

    def test_basic(self):
        with LocalTransport():
            pass


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import os

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

# Get the version number
aiida_folder = os.path.split(os.path.abspath(__file__))[0]
# This was importing dateutil, which is not available by default
#fname = os.path.join(aiida_folder, 'aiida', '__init__.py')
#with open(fname) as aiida_init:
#    ns = {}
#    exec (aiida_init.read(), ns)
#    aiida_version = ns['__version__']

if os.path.exists('CHANGELOG_EPFL.txt'):
    # EPFL version
    aiida_name = 'aiida-epfl'
    aiida_license = 'MIT and EPFL licenses, see LICENSE.txt'
else:
    aiida_name = 'aiida-core'
    aiida_license = 'MIT license, see LICENSE.txt'

bin_folder = os.path.join(aiida_folder, 'bin')

with open(os.path.join(aiida_folder, 'requirements.txt'), 'r') as rfile:
     requirements = rfile.read().splitlines()

setup(
    name=aiida_name,
    url='http://www.aiida.net/',
    license=aiida_license,
    version=__version__,
    install_requires=requirements,
    packages=find_packages(),
    scripts=[os.path.join(bin_folder, f) for f in os.listdir(bin_folder)
             if not os.path.isdir(os.path.join(bin_folder, f))],
    long_description=open(os.path.join(aiida_folder, 'README.rst')).read(),
)



# -*- coding: utf-8 -*-
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

from aiida.workflows2.wf import wf
from aiida.orm.data.simple import Int
from aiida.workflows2.db_types import to_db_type, SimpleData

@wf
def sum(a, b):
    return to_db_type(a.value + b.value)


@wf
def prod(a, b):
    return to_db_type(a.value * b.value)


@wf
def add_multiply_wf(a, b, c):
    return prod(sum(a, b), c)


if __name__ == '__main__':
    two = Int(2)
    three = Int(3)
    four = Int(4)

    print "WORKFUNCTION:"

    simpledata = add_multiply_wf(two, three, four)
    print "output pk:", simpledata.pk
    print "output value:", simpledata.value



# -*- coding: utf-8 -*-
import logging
import sys

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

aiidalogger = logging.getLogger("aiida")
# aiidalogger.addHandler(logging.StreamHandler(sys.stderr))
#FORMAT = '[%(name)s@%(levelname)s] %(message)s'
#logging.basicConfig(format=FORMAT)



# -*- coding: utf-8 -*-
import unittest

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class TestNodeNumberJobResource(unittest.TestCase):
    def test_init(self):
        """
        Test the __init__ of the NodeNumberJobResource class
        """
        from aiida.scheduler.datastructures import NodeNumberJobResource

        # No empty initialization
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource()

        # Missing required field
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_machines=1)
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=1)
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(tot_num_mpiprocs=1)

        # Standard info
        jr = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8)
        self.assertEquals(jr.num_machines, 2)
        self.assertEquals(jr.num_mpiprocs_per_machine, 8)
        self.assertEquals(jr.get_tot_num_mpiprocs(), 16)
        # redundant but consistent information
        jr = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8, tot_num_mpiprocs=16)
        self.assertEquals(jr.num_machines, 2)
        self.assertEquals(jr.num_mpiprocs_per_machine, 8)
        self.assertEquals(jr.get_tot_num_mpiprocs(), 16)
        # other equivalent ways of specifying the information
        jr = NodeNumberJobResource(num_mpiprocs_per_machine=8, tot_num_mpiprocs=16)
        self.assertEquals(jr.num_machines, 2)
        self.assertEquals(jr.num_mpiprocs_per_machine, 8)
        self.assertEquals(jr.get_tot_num_mpiprocs(), 16)
        # other equivalent ways of specifying the information
        jr = NodeNumberJobResource(num_machines=2, tot_num_mpiprocs=16)
        self.assertEquals(jr.num_machines, 2)
        self.assertEquals(jr.num_mpiprocs_per_machine, 8)
        self.assertEquals(jr.get_tot_num_mpiprocs(), 16)

        # wrong field name
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8, wrong_name=16)

        # Examples of wrong informaton (e.g., number of machines or of nodes < 0
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=0, num_mpiprocs_per_machine=8)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=1, num_mpiprocs_per_machine=0)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=1, tot_num_mpiprocs=0)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=1, tot_num_mpiprocs=0)

        # Examples of inconsistent information
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=8, num_machines=2, tot_num_mpiprocs=32)

        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=8, tot_num_mpiprocs=15)
        
        


# -*- coding: utf-8 -*-
import sys

from aiida.backends.utils import load_dbenv
from aiida.cmdline.baseclass import VerdiCommand

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class Export(VerdiCommand):
    """
    Export nodes and group of nodes

    This command allows to export to file nodes and group of nodes, for backup
    purposes or to share data with collaborators.
    Call this command with the '-h' option for some documentation of its usage.
    """

    def run(self, *args):
        load_dbenv()

        import argparse

        from aiida.common.exceptions import NotExistent
        from aiida.backends.djsite.db import models
        from aiida.orm import Group
        from aiida.orm.importexport import export, export_zip

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Export data from the DB.')
        parser.add_argument('-c', '--computers', nargs='+', type=int, metavar="PK",
                            help="Export the given computers")
        parser.add_argument('-n', '--nodes', nargs='+', type=int, metavar="PK",
                            help="Export the given nodes")
        parser.add_argument('-g', '--groups', nargs='+', metavar="GROUPNAME",
                            help="Export all nodes in the given group(s), identified by name.",
                            type=str)
        parser.add_argument('-G', '--group_pks', nargs='+', metavar="PK",
                            help="Export all nodes in the given group(s), identified by pk.",
                            type=str)
        parser.add_argument('-P', '--no-parents',
                            dest='no_parents', action='store_true',
                            help="Store only the nodes that are explicitly given, without exporting the parents")
        parser.set_defaults(no_parents=False)
        parser.add_argument('-O', '--no-calc-outputs',
                            dest='no_calc_outputs', action='store_true',
                            help="If a calculation is included in the list of nodes to export, do not export its outputs")
        parser.set_defaults(no_calc_outputs=False)
        parser.add_argument('-y', '--overwrite',
                            dest='overwrite', action='store_true',
                            help="Overwrite the output file, if it exists")
        parser.set_defaults(overwrite=False)

        zipsubgroup = parser.add_mutually_exclusive_group()
        zipsubgroup.add_argument('-z', '--zipfile-compressed',
                            dest='zipfilec', action='store_true',
                            help="Store as zip file (experimental, should be faster")
        zipsubgroup.add_argument('-Z', '--zipfile-uncompressed',
                            dest='zipfileu', action='store_true',
                            help="Store as uncompressed zip file (experimental, should be faster")
        parser.set_defaults(zipfilec=False)
        parser.set_defaults(zipfileu=False)

        parser.add_argument('output_file', type=str,
                            help='The output file name for the export file')

        parsed_args = parser.parse_args(args)

        if parsed_args.nodes is None:
            node_pk_list = []
        else:
            node_pk_list = parsed_args.nodes

        groups_list = []

        if parsed_args.groups is not None:
            for group_name in parsed_args.groups:
                try:
                    group = Group.get_from_string(group_name)
                except (ValueError, NotExistent) as e:
                    print >> sys.stderr, e.message
                    sys.exit(1)
                node_pk_list += group.dbgroup.dbnodes.values_list('pk', flat=True)
                groups_list.append(group.dbgroup)
                
        if parsed_args.group_pks is not None:
            for group_pk in parsed_args.group_pks:
                try:
                    group = Group.get(pk=group_pk)
                except (ValueError, NotExistent) as e:
                    print >> sys.stderr, e.message
                    sys.exit(1)
                node_pk_list += group.dbgroup.dbnodes.values_list('pk', flat=True)
                groups_list.append(group.dbgroup)
        
        node_pk_list = set(node_pk_list)

        node_list = list(
            models.DbNode.objects.filter(pk__in=node_pk_list))
        missing_nodes = node_pk_list.difference(_.pk for _ in node_list)
        for pk in missing_nodes:
            print >> sys.stderr, ("WARNING! Node with pk= {} "
                                  "not found, skipping.".format(pk))
        if parsed_args.computers is not None:
            computer_list = list(models.DbComputer.objects.filter(
                pk__in=parsed_args.computers))
            missing_computers = set(parsed_args.computers).difference(
                _.pk for _ in computer_list)
            for pk in missing_computers:
                print >> sys.stderr, ("WARNING! Computer with pk= {} "
                                      "not found, skipping.".format(pk))
        else:
            computer_list = []

        what_list = node_list + computer_list + groups_list

        export_function = export
        additional_kwargs = {}
        if parsed_args.zipfileu:
            export_function = export_zip
            additional_kwargs.update({"use_compression": False})
        elif parsed_args.zipfilec:
            export_function = export_zip
            additional_kwargs.update({"use_compression": True})
        try:
            export_function(what=what_list,
                   also_parents=not parsed_args.no_parents,
                   also_calc_outputs=not parsed_args.no_calc_outputs,
                   outfile=parsed_args.output_file,
                   overwrite=parsed_args.overwrite,**additional_kwargs)
        except IOError as e:
            print >> sys.stderr, "IOError: {}".format(e.message)
            sys.exit(1)


    def complete(self, subargs_idx, subargs):
        return ""



# -*- coding: utf-8 -*-
import importlib

import aiida.common
from aiida.common.exceptions import MissingPluginError

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

logger = aiida.common.aiidalogger.getChild('pluginloader')


def from_type_to_pluginclassname(typestr):
    """
    Return the string to pass to the load_plugin function, starting from
    the 'type' field of a Node.
    """
    # Fix for base class
    from aiida.common.exceptions import DbContentError
    if typestr == "":
        typestr = "node.Node."
    if not typestr.endswith("."):
        raise DbContentError("The type name '{}' is not valid!".format(
            typestr))
    return typestr[:-1]  # Strip final dot

def get_query_type_string(plugin_type_string):
    """
    Receives a plugin_type_string, an attribute of subclasses of Node.
    Checks whether it is a valid type_string and manipulates the string
    to return a string that in a query returns all instances of a class and
    all instances of subclasses.

    :param str plugin_type_string: The plugin_type_string

    :returns: the query_type_string

    """
    from aiida.common.exceptions import DbContentError, InputValidationError
    if not isinstance(plugin_type_string, basestring):
        raise InputValidationError("You have to pass as argument")
    # First case, an empty string for Node:
    if plugin_type_string == '':
        query_type_string = ''
    # Anything else should have baseclass.Class., so at least 2 dots 
    # and end with a dot:
    elif not(plugin_type_string.endswith('.')) or plugin_type_string.count('.') == 1:
        raise DbContentError(
                "The type name '{}' is not valid!".format(plugin_type_string)
            )
    else:
        query_type_string = '{}.'.format('.'.join(plugin_type_string.split('.')[:-2]))
    return query_type_string


def get_class_typestring(type_string):
    """
    Given the type string, return three strings: the first one is
    one of the first-level classes that the Node can be:
    "node", "calculation", "code", "data".
    The second string is the one that can be passed to the DataFactory or
    CalculationFactory (or an empty string for nodes and codes);
    the third one is the name of the python class that would be loaded.
    """
    from aiida.common.exceptions import DbContentError

    if type_string == "":
        return ("node", "")
    else:
        pieces = type_string.split('.')
        if pieces[-1]:
            raise DbContentError("The type string does not end with a dot")
        if len(pieces) < 3:
            raise DbContentError("Not enough parts in the type string")
        return pieces[0], ".".join(pieces[1:-2]), pieces[-2]


def _existing_plugins_with_module(base_class, plugins_module_path,
                                  pkgname, basename, max_depth, suffix=None):
    """
        Recursive function to return the existing plugins within a given module.

        :param base_class: Identify all subclasses of the base_class
        :param plugins_module_path: The path to the folder with the plugins
        :param pkgname: The name of the package in which you want to search
        :param basename: The basename of the plugin (sub)class. See also documentation
            of ``find_module``.
        :param max_depth: Maximum depth (of nested modules) to be used when
            looking for plugins
        :param suffix: The suffix that is appended to the basename when looking
            for the (sub)class name. If not provided (or None), use the base
            class name.
        :return: a list of valid strings that can be used using a Factory or with
            load_plugin.
    """
    import pkgutil
    import os

    if max_depth == 0:
        return []
    else:
        retlist = _find_module(base_class, pkgname, basename, suffix)

        for _, name, ismod in pkgutil.walk_packages([plugins_module_path]):
            if ismod:
                retlist += _existing_plugins_with_module(
                    base_class, os.path.join(plugins_module_path, name),
                    "{}.{}".format(pkgname, name),
                    "{}.{}".format(basename, name) if basename else name,
                    max_depth - 1, suffix=suffix)

            # This has to be done anyway, for classes in the __init__ file.
            this_pkgname = "{}.{}".format(pkgname, name)
            this_basename = "{}.{}".format(basename, name) if basename else name

            retlist += _find_module(base_class, this_pkgname, this_basename, suffix)

        return list(set(retlist))


def _find_module(base_class, pkgname, this_basename, suffix=None):
    """
    Given a base class object, looks for its subclasses inside the package
    with name pkgname (must be importable), and prepends to the class name
    the string 'this_basename'.

    If the name of the class complies with the syntax
    AaaBbb
    where Aaa is the capitalized name of the containing module (aaa), and
    Bbb is base_class.__name__, then only 'aaa' is returned instead of
    'aaa.AaaBbb', to have a shorter name that is anyway accepted by the *Factory
    functions. If suffix is provided, this is used for comparison (the 'Bbb'
    string) rather than the base class name)

    :param base_class: Identify all subclasses of the base_class
    :param pkgname: The name of the package in which you want to search
    :param basename: The basename of the plugin (sub)class. See also documentation
        of ``find_module``.
    :param suffix: The suffix that is appended to the basename when looking
        for the (sub)class name. If not provided (or None), use the base
        class name.
    :return: a list of valid strings, acceptable by the *Factory functions.
       Does not return the class itself.
    """
    import inspect

    retlist = []

    # print ' '*(5-max_depth), '>', pkgname
    #print ' '*(5-max_depth), ' ', this_basename

    pkg = importlib.import_module(pkgname)
    for k, v in pkg.__dict__.iteritems():
        if (inspect.isclass(v) and  # A class
                    v != base_class and  # Not the class itself
                issubclass(v, base_class) and  # a proper subclass
                    pkgname == v.__module__):  # We are importing it from its
            # module: avoid to import it
            # from another module, if it
            # was simply imported there
            # Try to return the shorter name if the subclass name
            # has the correct pattern, as expected by the Factory
            # functions
            if suffix is None:
                actual_suffix = base_class.__name__
            else:
                actual_suffix = suffix

            if k == "{}{}".format(
                    pkgname.rpartition('.')[2].capitalize(),
                    actual_suffix):
                retlist.append(this_basename)
            else:
                retlist.append(
                    ("{}.{}".format(this_basename, k) if this_basename
                     else k))
                #print ' '*(5-max_depth), ' ->', "{}.{}".format(this_basename, k)
    return retlist


def existing_plugins(base_class, plugins_module_name, max_depth=5, suffix=None):
    """
    Return a list of strings of valid plugins.


    :param base_class: Identify all subclasses of the base_class
    :param plugins_module_name: a string with the full module name separated
        with dots that points to the folder with plugins.
        It must be importable by python.
    :param max_depth: Maximum depth (of nested modules) to be used when
            looking for plugins
    :param suffix: The suffix that is appended to the basename when looking
        for the (sub)class name. If not provided (or None), use the base
        class name.
    :return: a list of valid strings that can be used using a Factory or with
        load_plugin.
    """
    try:
        pluginmod = importlib.import_module(plugins_module_name)
    except ImportError:
        raise MissingPluginError("Unable to load the plugin module {}".format(
            plugins_module_name))

    return _existing_plugins_with_module(base_class,
                                         pluginmod.__path__[0],
                                         plugins_module_name,
                                         "",
                                         max_depth, suffix)


def load_plugin(base_class, plugins_module, plugin_type):
    """
    Load a specific plugin for the given base class.

    This is general and works for any plugin used in AiiDA.

    NOTE: actually, now plugins_module and plugin_type are joined with a dot,
        and the plugin is retrieved splitting using the last dot of the resulting
        string.
    TODO: understand if it is probably better to join the two parameters above
        to a single one.

    Args:
        base_class
            the abstract base class of the plugin.
        plugins_module
            a string with the full module name separated with dots
            that points to the folder with plugins. It must be importable by python.
        plugin_type
            the name of the plugin.

    Return:
        the class of the required plugin.

    Raise:
        MissingPluginError if the plugin cannot be loaded

    Example:
       plugin_class = load_plugin(
           aiida.transport.Transport,'aiida.transport.plugins','ssh.SshTransport')

       and plugin_class will be the class 'aiida.transport.plugins.ssh.SshTransport'
    """

    module_name = ".".join([plugins_module, plugin_type])
    real_plugin_module, plugin_name = module_name.rsplit('.', 1)


    try:
        pluginmod = importlib.import_module(real_plugin_module)
    except ImportError:
        raise MissingPluginError("Unable to load the plugin module {}".format(
            real_plugin_module))

    try:
        pluginclass = pluginmod.__dict__[plugin_name]
    except KeyError:
        raise MissingPluginError("Unable to load the class {} within {}".format(
            plugin_name, real_plugin_module))

    try:
        if issubclass(pluginclass, base_class):
            return pluginclass
        else:
            # Quick way of going into the except case
            err_msg = "{} is not a subclass of {}".format(
                module_name, base_class.__name__)
            raise MissingPluginError(err_msg)
    except TypeError:
        # This happens when we pass a non-class to issubclass;
        err_msg = "{} is not a class".format(
            module_name)
        raise MissingPluginError(err_msg)


def BaseFactory(module, base_class, base_modname, suffix=None):
    """
    Return a given subclass of Calculation, loading the correct plugin.

    :example: If `module='quantumespresso.pw'`, `base_class=JobCalculation`,
      `base_modname = 'aiida.orm.calculation.job'`, and `suffix='Calculation'`,
      the code will first look for a pw subclass of JobCalculation
      inside the quantumespresso module. Lacking such a class, it will try to look
      for a 'PwCalculation' inside the quantumespresso.pw module.
      In the latter case, the plugin class must have a specific name and be
      located in a specific file:
      if for instance plugin_name == 'ssh' and base_class.__name__ == 'Transport',
      then there must be a class named 'SshTransport' which is a subclass of base_class
      in a file 'ssh.py' in the plugins_module folder.
      To create the class name to look for, the code will attach the string
      passed in the base_modname (after the last dot) and the suffix parameter,
      if passed, with the proper CamelCase capitalization. If suffix is not
      passed, the default suffix that is used is the base_class class name.

    :param module: a string with the module of the plugin to load, e.g.
          'quantumespresso.pw'.
    :param base_class: a base class from which the returned class should inherit.
           e.g.: JobCalculation
    :param base_modname: a basic module name, under which the module should be
            found. E.g., 'aiida.orm.calculation.job'.
    :param suffix: If specified, the suffix that the class name will have.
      By default, use the name of the base_class.
    """
    try:
        return load_plugin(base_class, base_modname, module)
    except MissingPluginError as e1:
        # Automatically add subclass name and try again
        if suffix is None:
            actual_suffix = base_class.__name__
        else:
            actual_suffix = suffix
        mname = module.rpartition('.')[2].capitalize() + actual_suffix
        new_module = module + '.' + mname
        try:
            return load_plugin(base_class, base_modname, new_module)
        except MissingPluginError as e2:
            err_msg = ("Neither {} or {} could be loaded from {}. "
                       "Error messages were: '{}', '{}'").format(
                module, new_module, base_modname, e1, e2)
            raise MissingPluginError(err_msg)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import getpass
import os

import aiida.common.setup as setup

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class Migration(object):
    """
    This class converts the configuration from the AiiDA version with one
    backend (Django) as it is described in commit e048939 to the implementation
    that supports DJango and SQLAlchemy.

    It is assumed that it is already (correctly) defined the place that the
    various configuration files are found in the aiida.common.setup.
    """

    # Profile key
    _profiles_key = "profiles"

    def adding_backend(self):
        """
        Gets the main AiiDA configuration and searches if there is a backend
        defined. If there isn't any then Django is added.
        """
        # Get the available configuration
        conf = setup.get_config()

        # Identifying all the available profiles
        if self._profiles_key in conf.keys():
            profiles = conf[self._profiles_key]
            p_keys = profiles.keys()
            # For every profile
            for p_key in p_keys:
                # get the
                curr_profile = profiles[p_key]

                # Check if there is a specific backend in the profile
                # and if not, add Django as backend
                if setup.aiidadb_backend_key not in curr_profile.keys():
                    curr_profile[setup.aiidadb_backend_key] = \
                        setup.aiidadb_backend_value_django

        # Returning the configuration
        return conf

    def perform_migration(self):
        # Backup the previous config
        setup.backup_config()
        # Get the AiiDA directory path
        aiida_directory = os.path.expanduser(setup.AIIDA_CONFIG_FOLDER)
        # Construct the daemon directory path
        daemon_dir = os.path.join(aiida_directory, setup.DAEMON_SUBDIR)
        # Construct the log directory path
        log_dir = os.path.join(aiida_directory, setup.LOG_SUBDIR)
        # Update the configuration if needed
        confs = self.adding_backend()
        # Store the configuration
        setup.store_config(confs)
        # Update the daemon directory
        setup.install_daemon_files(aiida_directory, daemon_dir, log_dir,
                             getpass.getuser())

if __name__ == '__main__':
    Migration().perform_migration()
    print("Migration finished.")



# -*- coding: utf-8 -*-

import copy

from django.db import IntegrityError, transaction
from django.db.models import F
from django.core.exceptions import ObjectDoesNotExist

from aiida.orm.implementation.general.node import AbstractNode, _NO_DEFAULT
from aiida.common.exceptions import (InternalError, ModificationNotAllowed,
                                     NotExistent, UniquenessError)
from aiida.common.utils import get_new_uuid
from aiida.common.folders import RepositoryFolder
from aiida.common.links import LinkType
from aiida.common.lang import override

from aiida.backends.djsite.utils import get_automatic_user
from aiida.backends.djsite.db.models import DbLink

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class Node(AbstractNode):
    @classmethod
    def get_subclass_from_uuid(cls, uuid):
        from aiida.backends.djsite.db.models import DbNode
        try:
            node = DbNode.objects.get(uuid=uuid).get_aiida_class()
        except ObjectDoesNotExist:
            raise NotExistent("No entry with UUID={} found".format(uuid))
        if not isinstance(node, cls):
            raise NotExistent("UUID={} is not an instance of {}".format(
                uuid, cls.__name__))
        return node

    @classmethod
    def get_subclass_from_pk(cls, pk):
        from aiida.backends.djsite.db.models import DbNode
        try:
            node = DbNode.objects.get(pk=pk).get_aiida_class()
        except ObjectDoesNotExist:
            raise NotExistent("No entry with pk= {} found".format(pk))
        if not isinstance(node, cls):
            raise NotExistent("pk= {} is not an instance of {}".format(
                pk, cls.__name__))
        return node

    def __int__(self):
        if self._to_be_stored:
            return None
        else:
            return self._dbnode.pk

    def __init__(self, **kwargs):
        from aiida.backends.djsite.db.models import DbNode
        super(Node, self).__init__()

        self._temp_folder = None

        dbnode = kwargs.pop('dbnode', None)

        # Set the internal parameters
        # Can be redefined in the subclasses
        self._init_internal_params()

        if dbnode is not None:
            if not isinstance(dbnode, DbNode):
                raise TypeError("dbnode is not a DbNode instance")
            if dbnode.pk is None:
                raise ValueError("If cannot load an aiida.orm.Node instance "
                                 "from an unsaved Django DbNode object.")
            if kwargs:
                raise ValueError("If you pass a dbnode, you cannot pass any "
                                 "further parameter")

            # If I am loading, I cannot modify it
            self._to_be_stored = False

            self._dbnode = dbnode

            # If this is changed, fix also the importer
            self._repo_folder = RepositoryFolder(section=self._section_name,
                                                 uuid=self._dbnode.uuid)

        # NO VALIDATION ON __init__ BY DEFAULT, IT IS TOO SLOW SINCE IT OFTEN
        # REQUIRES MULTIPLE DB HITS
        # try:
        #                # Note: the validation often requires to load at least one
        #                # attribute, and therefore it will take a lot of time
        #                # because it has to cache every attribute.
        #                self._validate()
        #            except ValidationError as e:
        #                raise DbContentError("The data in the DB with UUID={} is not "
        #                                     "valid for class {}: {}".format(
        #                    uuid, self.__class__.__name__, e.message))
        else:
            # TODO: allow to get the user from the parameters
            user = get_automatic_user()
            self._dbnode = DbNode(user=user,
                                  uuid=get_new_uuid(),
                                  type=self._plugin_type_string)

            self._to_be_stored = True

            # As creating the temp folder may require some time on slow
            # filesystems, we defer its creation
            self._temp_folder = None
            # Used only before the first save
            self._attrs_cache = {}
            # If this is changed, fix also the importer
            self._repo_folder = RepositoryFolder(section=self._section_name,
                                                 uuid=self.uuid)

            # Automatically set all *other* attributes, if possible, otherwise
            # stop
            self._set_with_defaults(**kwargs)

    @classmethod
    def query(cls, *args, **kwargs):
        from aiida.backends.djsite.db.models import DbNode
        if cls._plugin_type_string:
            if not cls._plugin_type_string.endswith('.'):
                raise InternalError("The plugin type string does not "
                                    "finish with a dot??")

            # If it is 'calculation.Calculation.', we want to filter
            # for things that start with 'calculation.' and so on
            plug_type = cls._plugin_type_string

            # Remove the implementation.django or sqla part.
            if plug_type.startswith('implementation.'):
                plug_type = '.'.join(plug_type.split('.')[2:])
            pre, sep, _ = plug_type[:-1].rpartition('.')
            superclass_string = "".join([pre, sep])
            return DbNode.aiidaobjects.filter(
                *args, type__startswith=superclass_string, **kwargs)
        else:
            # Base Node class, with empty string
            return DbNode.aiidaobjects.filter(*args, **kwargs)

    def _update_db_label_field(self, field_value):
        self.dbnode.label = field_value
        if not self._to_be_stored:
            with transaction.commit_on_success():
                self._dbnode.save()
                self._increment_version_number_db()

    def _update_db_description_field(self, field_value):
        self.dbnode.description = field_value
        if not self._to_be_stored:
            with transaction.commit_on_success():
                self._dbnode.save()
                self._increment_version_number_db()

    def _replace_dblink_from(self, src, label, link_type):
        try:
            self._add_dblink_from(src, label, link_type)
        except UniquenessError:
            # I have to replace the link; I do it within a transaction
            with transaction.commit_on_success():
                self._remove_dblink_from(label)
                self._add_dblink_from(src, label, link_type)

    def _remove_dblink_from(self, label):
        DbLink.objects.filter(output=self.dbnode, label=label).delete()

    def _add_dblink_from(self, src, label=None, link_type=LinkType.UNSPECIFIED):
        from aiida.backends.djsite.db.models import DbPath
        if not isinstance(src, Node):
            raise ValueError("src must be a Node instance")
        if self.uuid == src.uuid:
            raise ValueError("Cannot link to itself")

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "Cannot call the internal _add_dblink_from if the "
                "destination node is not stored")
        if src._to_be_stored:
            raise ModificationNotAllowed(
                "Cannot call the internal _add_dblink_from if the "
                "source node is not stored")

        if link_type is LinkType.CREATE or link_type is LinkType.INPUT:
            # Check for cycles. This works if the transitive closure is enabled; if it
            # isn't, this test will never fail, but then having a circular link is not
            # meaningful but does not pose a huge threat
            #
            # I am linking src->self; a loop would be created if a DbPath exists already
            # in the TC table from self to src
            if len(DbPath.objects.filter(parent=self.dbnode, child=src.dbnode)) > 0:
                raise ValueError(
                    "The link you are attempting to create would generate a loop")

        if label is None:
            autolabel_idx = 1

            existing_from_autolabels = list(DbLink.objects.filter(
                output=self.dbnode,
                label__startswith="link_").values_list('label', flat=True))
            while "link_{}".format(autolabel_idx) in existing_from_autolabels:
                autolabel_idx += 1

            safety_counter = 0
            while True:
                safety_counter += 1
                if safety_counter > 100:
                    # Well, if you have more than 100 concurrent addings
                    # to the same node, you are clearly doing something wrong...
                    raise InternalError("Hey! We found more than 100 concurrent"
                                        " adds of links "
                                        "to the same nodes! Are you really doing that??")
                try:
                    self._do_create_link(src, "link_{}".format(autolabel_idx), link_type)
                    break
                except UniquenessError:
                    # Retry loop until you find a new loop
                    autolabel_idx += 1
        else:
            self._do_create_link(src, label, link_type)

    def _do_create_link(self, src, label, link_type):
        sid = None
        try:
            # transactions are needed here for Postgresql:
            # https://docs.djangoproject.com/en/1.5/topics/db/transactions/#handling-exceptions-within-postgresql-transactions
            sid = transaction.savepoint()
            DbLink.objects.create(input=src.dbnode, output=self.dbnode,
                                  label=label, type=link_type.value)
            transaction.savepoint_commit(sid)
        except IntegrityError as e:
            transaction.savepoint_rollback(sid)
            raise UniquenessError("There is already a link with the same "
                                  "name (raw message was {})"
                                  "".format(e.message))

    def get_inputs(self, node_type=None, also_labels=False, only_in_db=False,
                   link_type=None):
        from aiida.backends.djsite.db.models import DbLink

        link_filter = {'output': self.dbnode}
        if link_type is not None:
            link_filter['type'] = link_type.value
        inputs_list = [(i.label, i.input.get_aiida_class()) for i in
                       DbLink.objects.filter(**link_filter).distinct()]

        if not only_in_db:
            # Needed for the check
            input_list_keys = [i[0] for i in inputs_list]

            for label, v in self._inputlinks_cache.iteritems():
                src = v[0]
                if label in input_list_keys:
                    raise InternalError("There exist a link with the same name "
                                        "'{}' both in the DB and in the internal "
                                        "cache for node pk= {}!".format(label, self.pk))
                inputs_list.append((label, src))

        if node_type is None:
            filtered_list = inputs_list
        else:
            filtered_list = [i for i in inputs_list if isinstance(i[1], node_type)]

        if also_labels:
            return list(filtered_list)
        else:
            return [i[1] for i in filtered_list]

    @override
    def get_outputs(self, type=None, also_labels=False, link_type=None):
        from aiida.backends.djsite.db.models import DbLink

        link_filter = {'input': self.dbnode}
        if link_type is not None:
            link_filter['type'] = link_type.value
        outputs_list = ((i.label, i.output.get_aiida_class()) for i in
                        DbLink.objects.filter(**link_filter).distinct())

        if type is None:
            if also_labels:
                return list(outputs_list)
            else:
                return [i[1] for i in outputs_list]
        else:
            filtered_list = (i for i in outputs_list if isinstance(i[1], type))
            if also_labels:
                return list(filtered_list)
            else:
                return [i[1] for i in filtered_list]

    def set_computer(self, computer):
        from aiida.backends.djsite.db.models import DbComputer
        if self._to_be_stored:
            self.dbnode.dbcomputer = DbComputer.get_dbcomputer(computer)
        else:
            raise ModificationNotAllowed(
                "Node with uuid={} was already stored".format(self.uuid))

    def _set_attr(self, key, value):
        from aiida.backends.djsite.db.models import DbAttribute
        DbAttribute.validate_key(key)

        if self._to_be_stored:
            self._attrs_cache[key] = copy.deepcopy(value)
        else:
            DbAttribute.set_value_for_node(self.dbnode, key, value)
            self._increment_version_number_db()

    def _del_attr(self, key):
        from aiida.backends.djsite.db.models import DbAttribute
        if self._to_be_stored:
            try:
                del self._attrs_cache[key]
            except KeyError:
                raise AttributeError(
                    "DbAttribute {} does not exist".format(key))
        else:
            if not DbAttribute.has_key(self.dbnode, key):
                raise AttributeError("DbAttribute {} does not exist".format(
                    key))
            DbAttribute.del_value_for_node(self.dbnode, key)
            self._increment_version_number_db()

    def get_attr(self, key, default=_NO_DEFAULT):
        from aiida.backends.djsite.db.models import DbAttribute
        try:
            if self._to_be_stored:
                try:
                    return self._attrs_cache[key]
                except KeyError:
                    raise AttributeError(
                        "DbAttribute '{}' does not exist".format(key))
            else:
                return DbAttribute.get_value_for_node(
                    dbnode=self.dbnode, key=key)
        except AttributeError:
            if default is _NO_DEFAULT:
                raise
            return default

    def set_extra(self, key, value, exclusive=False):
        from aiida.backends.djsite.db.models import DbExtra
        DbExtra.validate_key(key)

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "The extras of a node can be set only after "
                "storing the node")
        DbExtra.set_value_for_node(self.dbnode, key, value,
                                   stop_if_existing=exclusive)
        self._increment_version_number_db()

    def set_extra_exclusive(self, key, value):
        from aiida.backends.djsite.db.models import DbExtra
        DbExtra.validate_key(key)

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "The extras of a node can be set only after "
                "storing the node")
        DbExtra.set_value_for_node(self.dbnode, key, value,
                                   stop_if_existing=True)
        self._increment_version_number_db()

    def get_extra(self, key, *args):
        from aiida.backends.djsite.db.models import DbExtra
        if len(args) > 1:
            raise ValueError("After the key name you can pass at most one"
                             "value, that is the default value to be used "
                             "if no extra is found.")

        try:
            if self._to_be_stored:
                raise AttributeError("DbExtra '{}' does not exist yet, the "
                                     "node is not stored".format(key))
            else:
                return DbExtra.get_value_for_node(dbnode=self.dbnode,
                                                  key=key)
        except AttributeError as e:
            try:
                return args[0]
            except IndexError:
                raise e

    def get_extras(self):
        return dict(self.iterextras())

    def del_extra(self, key):
        from aiida.backends.djsite.db.models import DbExtra
        if self._to_be_stored:
            raise ModificationNotAllowed(
                "The extras of a node can be set and deleted "
                "only after storing the node")
        if not DbExtra.has_key(self.dbnode, key):
            raise AttributeError("DbExtra {} does not exist".format(
                key))
        return DbExtra.del_value_for_node(self.dbnode, key)
        self._increment_version_number_db()

    def extras(self):
        from aiida.backends.djsite.db.models import DbExtra
        if self._to_be_stored:
            return
        else:
            extraslist = DbExtra.list_all_node_elements(self.dbnode)
        for e in extraslist:
            yield e.key

    def iterextras(self):
        from aiida.backends.djsite.db.models import DbExtra
        if self._to_be_stored:
            # If it is not stored yet, there are no extras that can be
            # added (in particular, we do not even have an ID to use!)
            # Return without value, meaning that this is an empty generator
            return
        else:
            extraslist = DbExtra.list_all_node_elements(self.dbnode)
            for e in extraslist:
                yield (e.key, e.getvalue())

    def iterattrs(self):
        from aiida.backends.djsite.db.models import DbAttribute
        # TODO: check what happens if someone stores the object while
        #        the iterator is being used!
        if self._to_be_stored:
            for k, v in self._attrs_cache.iteritems():
                yield (k, v)
        else:
            all_attrs = DbAttribute.get_all_values_for_node(self.dbnode)
            for attr in all_attrs:
                yield (attr, all_attrs[attr])

    def get_attrs(self):
        return dict(self.iterattrs())

    def attrs(self):
        from aiida.backends.djsite.db.models import DbAttribute
        # Note: I "duplicate" the code from iterattrs, rather than
        # calling iterattrs from here, because iterattrs is slow on each call
        # since it has to call .getvalue(). To improve!
        if self._to_be_stored:
            for k in self._attrs_cache.iterkeys():
                yield k
        else:
            attrlist = DbAttribute.list_all_node_elements(self.dbnode)
            for attr in attrlist:
                yield attr.key

    def add_comment(self, content, user=None):
        from aiida.backends.djsite.db.models import DbComment
        if self._to_be_stored:
            raise ModificationNotAllowed("Comments can be added only after "
                                         "storing the node")

        DbComment.objects.create(dbnode=self._dbnode,
                                 user=user,
                                 content=content)

    def get_comment_obj(self, id=None, user=None):
        from aiida.backends.djsite.db.models import DbComment
        import operator
        from django.db.models import Q
        query_list = []

        # If an id is specified then we add it to the query
        if id is not None:
            query_list.append(Q(pk=id))

        # If a user is specified then we add it to the query
        if user is not None:
            query_list.append(Q(user=user))

        dbcomments = DbComment.objects.filter(
            reduce(operator.and_, query_list))
        comments = []
        from aiida.orm.implementation.django.comment import Comment
        for dbcomment in dbcomments:
            comments.append(Comment(dbcomment=dbcomment))
        return comments

    def get_comments(self, pk=None):
        from aiida.backends.djsite.db.models import DbComment
        if pk is not None:
            try:
                correct = all([isinstance(_, int) for _ in pk])
                if not correct:
                    raise ValueError('pk must be an integer or a list of integers')
            except TypeError:
                if not isinstance(pk, int):
                    raise ValueError('pk must be an integer or a list of integers')
            return list(DbComment.objects.filter(
                dbnode=self._dbnode, pk=pk).order_by('pk').values(
                'pk', 'user__email', 'ctime', 'mtime', 'content'))

        return list(DbComment.objects.filter(dbnode=self._dbnode).order_by(
            'pk').values('pk', 'user__email', 'ctime', 'mtime', 'content'))

    def _get_dbcomments(self, pk=None):
        from aiida.backends.djsite.db.models import DbComment
        if pk is not None:
            try:
                correct = all([isinstance(_, int) for _ in pk])
                if not correct:
                    raise ValueError('pk must be an integer or a list of integers')
                return list(DbComment.objects.filter(dbnode=self._dbnode, pk__in=pk).order_by('pk'))
            except TypeError:
                if not isinstance(pk, int):
                    raise ValueError('pk must be an integer or a list of integers')
                return list(DbComment.objects.filter(dbnode=self._dbnode, pk=pk).order_by('pk'))

        return list(DbComment.objects.filter(dbnode=self._dbnode).order_by('pk'))

    def _update_comment(self, new_field, comment_pk, user):
        from aiida.backends.djsite.db.models import DbComment
        comment = list(DbComment.objects.filter(dbnode=self._dbnode,
                                                pk=comment_pk, user=user))[0]

        if not isinstance(new_field, basestring):
            raise ValueError("Non string comments are not accepted")

        if not comment:
            raise NotExistent("Found no comment for user {} and pk {}".format(
                user, comment_pk))

        comment.content = new_field
        comment.save()

    def _remove_comment(self, comment_pk, user):
        from aiida.backends.djsite.db.models import DbComment
        comment = DbComment.objects.filter(dbnode=self._dbnode, pk=comment_pk)[0]
        comment.delete()

    def _increment_version_number_db(self):
        from aiida.backends.djsite.db.models import DbNode
        # I increment the node number using a filter
        self._dbnode.nodeversion = F('nodeversion') + 1
        self._dbnode.save()

        # This reload internally the node of self._dbnode
        # Note: I have to reload the object (to have the right values in memory,
        # otherwise I only get the Django Field F object as a result!
        self._dbnode = DbNode.objects.get(pk=self._dbnode.pk)

    def copy(self):
        newobject = self.__class__()
        newobject.dbnode.type = self.dbnode.type  # Inherit type
        newobject.dbnode.label = self.dbnode.label  # Inherit label
        # TODO: add to the description the fact that this was a copy?
        newobject.dbnode.description = self.dbnode.description  # Inherit description
        newobject.dbnode.dbcomputer = self.dbnode.dbcomputer  # Inherit computer

        for k, v in self.iterattrs():
            newobject._set_attr(k, v)

        for path in self.get_folder_list():
            newobject.add_path(self.get_abs_path(path), path)

        return newobject

    @property
    def uuid(self):
        return unicode(self.dbnode.uuid)

    @property
    def pk(self):
        return self.dbnode.pk

    @property
    def id(self):
        return self.dbnode.id

    @property
    def dbnode(self):
        # I also update the internal _dbnode variable, if it was saved
        # from aiida.backends.djsite.db.models import DbNode
        #        if not self._to_be_stored:
        #            self._dbnode = DbNode.objects.get(pk=self._dbnode.pk)
        return self._dbnode

    def store_all(self, with_transaction=True):
        """
        Store the node, together with all input links, if cached, and also the
        linked nodes, if they were not stored yet.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """
        from django.db import transaction
        from aiida.common.utils import EmptyContextManager

        if with_transaction:
            context_man = transaction.commit_on_success()
        else:
            context_man = EmptyContextManager()

        if not self._to_be_stored:
            raise ModificationNotAllowed(
                "Node with pk= {} was already stored".format(self.pk))

        # For each parent, check that all its inputs are stored
        for label in self._inputlinks_cache:
            try:
                parent_node = self._inputlinks_cache[label][0]
                parent_node._check_are_parents_stored()
            except ModificationNotAllowed:
                raise ModificationNotAllowed(
                    "Parent node (UUID={}) has "
                    "unstored parents, cannot proceed (only direct parents "
                    "can be unstored and will be stored by store_all, not "
                    "grandparents or other ancestors".format(parent_node.uuid))

        with context_man:
            # Always without transaction: either it is the context_man here,
            # or it is managed outside
            self._store_input_nodes()
            self.store(with_transaction=False)
            self._store_cached_input_links(with_transaction=False)

        return self

    def _store_input_nodes(self):
        """
        Find all input nodes, and store them, checking that they do not
        have unstored inputs in turn.

        :note: this function stores all nodes without transactions; always
          call it from within a transaction!
        """
        if not self._to_be_stored:
            raise ModificationNotAllowed(
                "_store_input_nodes can be called only if the node is "
                "unstored (node {} is stored, instead)".format(self.pk))

        for label in self._inputlinks_cache:
            parent = self._inputlinks_cache[label][0]
            if not parent.is_stored:
                parent.store(with_transaction=False)

    def _check_are_parents_stored(self):
        """
        Check if all parents are already stored, otherwise raise.

        :raise ModificationNotAllowed: if one of the input nodes in not already
          stored.
        """
        # Preliminary check to verify that inputs are stored already
        for label in self._inputlinks_cache:
            if not self._inputlinks_cache[label][0].is_stored:
                raise ModificationNotAllowed(
                    "Cannot store the input link '{}' because the "
                    "source node is not stored. Either store it first, "
                    "or call _store_input_links with the store_parents "
                    "parameter set to True".format(label))

    def _store_cached_input_links(self, with_transaction=True):
        """
        Store all input links that are in the local cache, transferring them
        to the DB.

        :note: This can be called only if all parents are already stored.

        :note: Links are stored only after the input nodes are stored. Moreover,
            link storage is done in a transaction, and if one of the links
            cannot be stored, an exception is raised and *all* links will remain
            in the cache.

        :note: This function can be called only after the node is stored.
           After that, it can be called multiple times, and nothing will be
           executed if no links are still in the cache.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """
        from django.db import transaction
        from aiida.common.utils import EmptyContextManager

        if with_transaction:
            context_man = transaction.commit_on_success()
        else:
            context_man = EmptyContextManager()

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "Node with pk= {} is not stored yet".format(self.pk))

        with context_man:
            # This raises if there is an unstored node.
            self._check_are_parents_stored()
            # I have to store only those links where the source is already
            # stored
            links_to_store = list(self._inputlinks_cache.keys())

            for label in links_to_store:
                src, link_type = self._inputlinks_cache[label]
                self._add_dblink_from(src, label, link_type)
            # If everything went smoothly, clear the entries from the cache.
            # I do it here because I delete them all at once if no error
            # occurred; otherwise, links will not be stored and I
            # should not delete them from the cache (but then an exception
            # would have been raised, and the following lines are not executed)
            self._inputlinks_cache.clear()

    def store(self, with_transaction=True):
        """
        Store a new node in the DB, also saving its repository directory
        and attributes.

        After being called attributes cannot be
        changed anymore! Instead, extras can be changed only AFTER calling
        this store() function.

        :note: After successful storage, those links that are in the cache, and
            for which also the parent node is already stored, will be
            automatically stored. The others will remain unstored.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """
        # TODO: This needs to be generalized, allowing for flexible methods
        # for storing data and its attributes.
        from django.db import transaction
        from aiida.common.utils import EmptyContextManager
        from aiida.common.exceptions import ValidationError
        from aiida.backends.djsite.db.models import DbAttribute
        import aiida.orm.autogroup

        if with_transaction:
            context_man = transaction.commit_on_success()
        else:
            context_man = EmptyContextManager()

        if self._to_be_stored:

            # As a first thing, I check if the data is valid
            self._validate()

            # Verify that parents are already stored. Raises if this is not
            # the case.
            self._check_are_parents_stored()

            # I save the corresponding django entry
            # I set the folder
            # NOTE: I first store the files, then only if this is successful,
            # I store the DB entry. In this way,
            # I assume that if a node exists in the DB, its folder is in place.
            # On the other hand, periodically the user might need to run some
            # bookkeeping utility to check for lone folders.
            self._repository_folder.replace_with_folder(
                self._get_temp_folder().abspath, move=True, overwrite=True)

            # I do the transaction only during storage on DB to avoid timeout
            # problems, especially with SQLite
            try:
                with context_man:
                    # Save the row
                    self._dbnode.save()
                    # Save its attributes 'manually' without incrementing
                    # the version for each add.
                    DbAttribute.reset_values_for_node(self.dbnode,
                                                      attributes=self._attrs_cache,
                                                      with_transaction=False)
                    # This should not be used anymore: I delete it to
                    # possibly free memory
                    del self._attrs_cache

                    self._temp_folder = None
                    self._to_be_stored = False

                    # Here, I store those links that were in the cache and
                    # that are between stored nodes.
                    self._store_cached_input_links()

            # This is one of the few cases where it is ok to do a 'global'
            # except, also because I am re-raising the exception
            except:
                # I put back the files in the sandbox folder since the
                # transaction did not succeed
                self._get_temp_folder().replace_with_folder(
                    self._repository_folder.abspath, move=True, overwrite=True)
                raise

            # Set up autogrouping used be verdi run
            autogroup = aiida.orm.autogroup.current_autogroup
            grouptype = aiida.orm.autogroup.VERDIAUTOGROUP_TYPE
            if autogroup is not None:
                if not isinstance(autogroup, aiida.orm.autogroup.Autogroup):
                    raise ValidationError("current_autogroup is not an AiiDA Autogroup")
                if autogroup.is_to_be_grouped(self):
                    group_name = autogroup.get_group_name()
                    if group_name is not None:
                        from aiida.orm import Group

                        g = Group.get_or_create(name=group_name, type_string=grouptype)[0]
                        g.add_nodes(self)

        # This is useful because in this way I can do
        # n = Node().store()
        return self

    @property
    def has_children(self):
        from aiida.backends.djsite.db.models import DbPath
        childrens = DbPath.objects.filter(parent=self.pk)
        return False if not childrens else True

    @property
    def has_parents(self):
        from aiida.backends.djsite.db.models import DbPath
        parents = DbPath.objects.filter(child=self.pk)
        return False if not parents else True



# -*- coding: utf-8 -*-

import unittest
import functools
import shutil


from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker


from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.utils import (
    load_dbenv, is_dbenv_loaded, get_configured_user_email, get_automatic_user,
    install_tc, loads_json, dumps_json
)
from aiida.backends.sqlalchemy.models.base import Base
from aiida.backends.sqlalchemy.models.user import DbUser
from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.orm.computer import Computer

from aiida.common.setup import get_profile_config

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

Session = sessionmaker()


class SqlAlchemyTests(unittest.TestCase):

    # Specify the need to drop the table at the beginning of a test case
    drop_all = False

    @classmethod
    def setUpClass(cls):

        config = get_profile_config("tests")
        engine_url = ("postgresql://{AIIDADB_USER}:{AIIDADB_PASS}@"
                      "{AIIDADB_HOST}:{AIIDADB_PORT}/{AIIDADB_NAME}").format(**config)
        engine = create_engine(engine_url,
                               json_serializer=dumps_json,
                               json_deserializer=loads_json)

        cls.connection = engine.connect()

        session = Session(bind=cls.connection)
        sa.session = session

        if cls.drop_all:
            Base.metadata.drop_all(cls.connection)
        Base.metadata.create_all(cls.connection)
        install_tc(cls.connection)

        email = get_configured_user_email()

        has_user = DbUser.query.filter(DbUser.email==email).first()
        if not has_user:
            user = DbUser(email, "foo", "bar", "tests")
            sa.session.add(user)
            sa.session.commit()
            sa.session.expire_all()

        has_computer = DbComputer.query.filter(DbComputer.hostname == 'localhost').first()
        if not has_computer:
            computer = SqlAlchemyTests._create_computer()
            computer.store()

        session.close()

    @staticmethod
    def _create_computer(**kwargs):
        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        defaults.update(kwargs)
        return Computer(**defaults)

    @staticmethod
    def inject_computer(f):
        @functools.wraps(f)
        def dec(*args, **kwargs):
            computer = DbComputer.query.filter_by(name="localhost").first()
            args = list(args)
            args.insert(1, computer)
            return f(*args, **kwargs)

        return dec


    @classmethod
    def tearDownClass(cls):
        # Clean what we added before
        cls.connection.close()
        config = get_profile_config("tests")
        repo_dir = config["AIIDADB_REPOSITORY_URI"]
        # We only treat the case where its a folder
        if repo_dir.startswith("file://"):
            repo_dir = repo_dir.split("file://")[-1]
            try:
                shutil.rmtree(repo_dir)
            except OSError:
                # If the folder doesn't exist, we don't care
                pass

    def setUp(self):
        connec = self.__class__.connection
        self.trans = connec.begin()
        self.session = Session(bind=connec)
        sa.session = self.session

        dbcomputer = DbComputer.query.filter_by(name="localhost").first()
        self.computer = Computer(dbcomputer=dbcomputer)

        self.session.begin_nested()

        # then each time that SAVEPOINT ends, reopen it
        @event.listens_for(self.session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            if transaction.nested and not transaction._parent.nested:

                # ensure that state is expired the way
                # session.commit() at the top level normally does
                # (optional step)
                session.expire_all()

                session.begin_nested()

    def tearDown(self):
        self.session.rollback()
        self.session.close()
        self.trans.rollback()

